/*
 * Copyright (c) 2003 David E. Bakken, his research students, and Washington State University.
 * Please see the file LICENSE.pdf for more details on terms of use.
 */

package ZEN_IDL.be;

import ZEN_IDL.ast.*;
import ZEN_IDL.be.symbolTableUtils.*;
import ZEN_IDL.be.helperUtils.*;
import util.*;

import java.io.*;
import java.util.*;

/*********************************************************
* Generates stubs and skeletons.
* Generates files <interfaceName>POA.java and 
* _<interfaceName>Stub.java.
*
* @author Olav Haugan
*
* 14.10.01
*********************************************************/
public class MQCCodeGenerator
{
	private String safeName; /* The safe name of the interface */
	private String packagePath; /* The path of this package */
	private String packageStr; /* The package string used for ID */
	private String packageName;/* The short package name */
	private String interfaceName; /* The name of the interface */
	
	private MQCCodeGenHelper codeGenHelper; /* The code generator helper object */
	private MQCAbstractMarshalHelper marshaller; /* The marshal helper object */
	private MQCAbstractMarshalHelper regMarshaller; /* The marshal helper object */
	private MQCConfig config; /* The configuration options */
	
	/*********************************************************
	* Constructor. Initializes variables.
	* 
	*********************************************************/
	public MQCCodeGenerator(MQCConfig config)
	{
		codeGenHelper = new MQCCodeGenHelper(config);		
		this.config = config;
		safeName = null;
		packagePath = null;
		packageStr = null;
		packageName = null;
		interfaceName = null;
		if(config.isInline())
		{
                   /*
			//If we use MQCIOP we can do simpler marshaling/unmarshaling
			if(config.getClientProtocol().equals("MQCIOP"))
			{
				if((config.isForcingEndianess() == 0) && config.isHomogeneousHW()) //big endian
				{
					//Force big endian marshaling/unmarshaling
					marshaller = new MQCInline_MQCIOP_FEMarshalHelper(config, true);
				}
				else if(config.isForcingEndianess() == 1)
				{
					//Force big endian marshaling/unmarshaling
					marshaller = new MQCInline_MQCIOP_FEMarshalHelper(config, true);
				}
				else if(config.isForcingEndianess() == 2)
				{
					//Force little endian marshaling/unmarshaling
					marshaller = new MQCInline_MQCIOP_FEMarshalHelper(config, false);
				}
				else
				{
					//Normal marshaling/unmarshaling
					marshaller = new MQCInline_MQCIOP_MarshalHelper(config);
				}
			}
			else
			{
				marshaller = findMarshaller();
			}
                        */

			marshaller = findMarshaller();

			regMarshaller = findMarshaller();
		}
		else
		{
			marshaller = new MQCLibMarshalHelper(config);
		}
	}
	
	/*********************************************************
	* Finds the correct marshaller
	* 
	* @return The marshaller
	*
	*********************************************************/
	private MQCAbstractMarshalHelper findMarshaller()
	{
		MQCAbstractMarshalHelper marshal;
		if((config.isForcingEndianess() == 0) && config.isHomogeneousHW()) //big endian
		{
			//Force big endian marshaling/unmarshaling
			marshal = new MQCInlineFEMarshalHelper(config, true);
		}
		else if(config.isForcingEndianess() == 1)
		{
			//Force big endian marshaling/unmarshaling
			marshal = new MQCInlineFEMarshalHelper(config, true);
		}
		else if(config.isForcingEndianess() == 2)
		{
			//Force little endian marshaling/unmarshaling
			marshal = new MQCInlineFEMarshalHelper(config, false);
		}
		else
		{
			//Normal marshaling/unmarshaling
			marshal = new MQCInlineMarshalHelper(config);
		}
		return marshal;
	}
	
	/*********************************************************
	* Sets the safe name variable
	* 
	* @param String The full name of the interface
	*
	*********************************************************/
	public void setSafeName( String safeName )
	{
		this.safeName = safeName;
	}
	
	/*********************************************************
	* Sets the package path variable
	* 
	* @param String The package path
	*
	*********************************************************/
	public void setPackagePath( String packagePath )
	{
		this.packagePath = packagePath;
	}
	
	/*********************************************************
	* Sets the package string variable
	* 
	* @param String The package string
	*
	*********************************************************/
	public void setPackageStr( String packageStr )
	{
		this.packageStr = packageStr;
	}
	
	/*********************************************************
	* Sets the package name
	* 
	* @param String The package name
	*
	*********************************************************/
	public void setPackageName( String packageName )
	{
		this.packageName = packageName;
	}

	/*********************************************************
	* Sets the interface name
	* 
	* @param String The interface name
	*
	*********************************************************/
	public void setInterfaceName( String interfaceName )
	{
		this.interfaceName = interfaceName;
	}
	
	/*********************************************************
	* Generates the stub/proxy code.
	* This is the client side code.
	* 
	* @param Vector The methods belonging to this interface
	*
	*********************************************************/
	public void genStubCode( Vector interfaceMethods )
	{
		PrintWriter pw = codeGenHelper.makeFile(packagePath, "_" + interfaceName + "Stub");

		if(packagePath.length() != 0)
			pw.println(packageStr);
		pw.println();

		pw.println("public class " + "_" + interfaceName + "Stub extends mqc.ObjImpl implements " + interfaceName);
		pw.println("{");

		/* Generate the method stubs */
		genMethodStubs(pw, interfaceMethods);
		
		pw.println("};");
		pw.println("// End of file.");
		pw.flush();
		pw.close();
	}
	
	/*********************************************************
	* Generates the method headers and calls genStubBody()
	* 
	* @param PrintWriter The outputstream
	*
	* @param Vector The methods belonging to this interface
	*
	*********************************************************/
	private void genMethodStubs( PrintWriter pw, Vector interfaceMethods)
	{
		Enumeration e = interfaceMethods.elements();
		InterfaceMethod iMethod = null;
		Vector params = null;
			
		while( e.hasMoreElements() ) 
		{
			iMethod = (InterfaceMethod) e.nextElement();
			params = iMethod.paramVector;
			pw.print("\tpublic " + codeGenHelper.mapIDL2Java( iMethod.returnType ) + " " + iMethod.name + "( " );
			codeGenHelper.printParams(pw, params);
			pw.print( " )");

			Enumeration ex = iMethod.exceptionVector.elements();
			Vector elem = null;
			if(ex.hasMoreElements())
			{
				pw.print( " throws mqc.exception.UNKNOWN, ");
				while(ex.hasMoreElements())
				{
					elem = (Vector) ex.nextElement();
					Enumeration exceptionElement = elem.elements();					
					while(exceptionElement.hasMoreElements())
					{
						pw.print(exceptionElement.nextElement().toString());
						if(exceptionElement.hasMoreElements())
							pw.print(".");
					}
					if(ex.hasMoreElements())
						pw.print(", ");
				}
			}
			pw.println();
			
			/* Generate the stub bodies */
			if(config.isInline())
				genInlineStubBody(pw, iMethod);
			else
				genStubBody(pw, iMethod);
		}
	}

	/*********************************************************
	* Generates the method stub bodies.
	* 
	* @param PrintWriter The outputstream
	*
	* @param InterfaceMethod The method that is to be generated
	*
	*********************************************************/
	private void genStubBody( PrintWriter pw, InterfaceMethod iMethod )
	{
		pw.println("\t{");
		pw.println("\t\tmqc.CDRInputStream _input = null;");
		pw.println("\t\tmqc.CDROutputStream _output = null;");

		if(!iMethod.returnType.equals("void"))
		{
			pw.print("\t\t" + codeGenHelper.mapIDL2Java( iMethod.returnType ) + " _result = ");
			if(iMethod.returnType.equals("boolean"))
				pw.println("false;");
			else
				pw.println("0;");
		}
		
		MQCPrinter printer = new MQCPrinter();
		printer.setPW(pw);
		printer.setIndent(2);		
		if(config.isSystemException())
		{
			pw.println("\t\ttry");
			pw.println("\t\t{");			
			printer.setIndent(3);
		}
		
		int funNo = 0;
		if(iMethod.isOneWay)
		{
			if(config.isCompressing())
			{
				funNo = MQCCompress.compressMethod(iMethod.name);
				printer.println("_output = this._request( \"" + funNo + "\", false);");
			}
			else
				printer.println("_output = this._request( \"" + iMethod.name +"\", false);");
		}
		else
		{
			if(config.isCompressing())
			{
				funNo = MQCCompress.compressMethod(iMethod.name);
				printer.println("_output = this._request( \"" + funNo + "\", true);");
			}
			else
				printer.println("_output = this._request( \"" + iMethod.name +"\", true);");
		}
			
		/* Print the client stub marshalling code */
		marshaller.printMarshal( printer, iMethod, true );

		printer.println("_input = this._invoke(_output);" );

		/* Print the client stub un-marshalling code */
		marshaller.printUnMarshal( printer, iMethod, true );

		if(config.isSystemException())
		{
			printer.setIndent(2);
			printer.println("}");			
			printer.println("catch(mqc.exception.SystemException e)");			
			printer.println("{");			
			printer.setIndent(3);
			printer.println("System.out.println(e.toString());");						
			printer.setIndent(2);
			printer.println("}");			

			Enumeration e = iMethod.exceptionVector.elements();
			if(e.hasMoreElements())
			{
				printer.setIndent(2);
				printer.println("catch(mqc.exception.UserException e)");			
				printer.println("{");			
				printer.setIndent(3);
				printer.println("java.lang.String exceptionID = e.getID();");

				Vector elem = null;
				while(e.hasMoreElements())
				{
					elem = (Vector) e.nextElement();
					Enumeration exceptionElement = elem.elements();					
					printer.print("if(exceptionID.equals(");
					while(exceptionElement.hasMoreElements())
					{
						printer.nprint(exceptionElement.nextElement().toString());
						if(exceptionElement.hasMoreElements())
							printer.nprint(".");
					}
					printer.nprintln("Helper.id()))");
					printer.println("{");
					printer.setIndent(4);
					printer.println("mqc.CDRInputStream input = e.getInputStream();");

					exceptionElement = elem.elements();
					printer.print("throw ");
					while(exceptionElement.hasMoreElements())
					{
						printer.nprint(exceptionElement.nextElement().toString());
						if(exceptionElement.hasMoreElements())
							printer.nprint(".");
					}
					printer.nprintln("Helper.read(input);");
					printer.setIndent(3);
					printer.println("}");
				}
				printer.println("throw new mqc.exception.UNKNOWN(\"Unknown user exception: \" + exceptionID );" );
				printer.setIndent(2);
				printer.println("}");			
			}
		}			
		if(!iMethod.returnType.equals("void"))
			printer.println("return _result;");
		pw.println("\t}");
	}

	/*********************************************************
	* Generates the inline method stub bodies.
	* 
	* @param PrintWriter The outputstream
	*
	* @param InterfaceMethod The method that is to be generated
	*
	*********************************************************/
	private void genInlineStubBody( PrintWriter pw, InterfaceMethod iMethod )
	{
		pw.println("\t{");
		pw.println("\t\tbyte[] _msg = null;");

		if(!iMethod.returnType.equals("void"))
		{
			pw.print("\t\t" + codeGenHelper.mapIDL2Java( iMethod.returnType ) + " _result = ");
			if(iMethod.returnType.equals("boolean"))
				pw.println("false;");
			else
				pw.println("0;");
		}
		
		int funNo = 0;
		if(config.isCompressing())
		{
			funNo = MQCCompress.compressMethod(iMethod.name);
		}

		MQCPrinter printer = new MQCPrinter();
		printer.setPW(pw);
		printer.setIndent(2);		
		if(config.isSystemException())
		{
			printer.println("try");
			printer.println("{");			
			printer.setIndent(3);
		}
	
		
		int bufferSize = 0;
		if(config.getClientProtocol().equals("MQCIOP"))
		{
			//The exact number of bytes needed for data types itself
			if(config.isCompressing())
				bufferSize = codeGenHelper.getInPayload(iMethod, String.valueOf(funNo), Constants.MQC_SIZE_OF_HDR + Constants.MQC_SIZE_OF_REQ_HDR);
			else
				bufferSize = codeGenHelper.getInPayload(iMethod, iMethod.name, Constants.MQC_SIZE_OF_HDR + Constants.MQC_SIZE_OF_REQ_HDR);
		}
		else
		{
			//Estimation of the number of bytes needed for data types itself
			bufferSize = codeGenHelper.getPayload(iMethod);
		}
		
		printer.println("mqc.holders.IntHolder ptr = new mqc.holders.IntHolder();");
		if(iMethod.isOneWay)
		{
			if(config.isCompressing())
				printer.println("_msg = this._request(\"" + funNo + "\", " + bufferSize + ", false, ptr );");
			else
				printer.println("_msg = this._request(\"" + iMethod.name.trim() + "\", " + bufferSize + ", false, ptr );");
		}
		else
		{
			if(config.isCompressing())
				printer.println("_msg = this._request(\"" + funNo + "\", " + bufferSize + ", true, ptr );");
			else
				printer.println("_msg = this._request(\"" + iMethod.name.trim() + "\", " + bufferSize + ", true, ptr );");
		}

		/* Print the client stub marshalling code */
		marshaller.init();
		int index = marshaller.printMarshal( printer, iMethod, true );

                /*
		if(config.getClientProtocol().equals("MQCIOP"))
			printer.println("ptr.value = " + index + ";");
               */
		if(iMethod.isOneWay)
			printer.println("this._invoke(_msg, ptr, false);" );
		else
			printer.println("_msg = this._invoke(_msg, ptr, true);" );

		/* Print the client stub un-marshalling code */
		marshaller.printUnMarshal( printer, iMethod, true );

		if(config.isSystemException())
		{
			printer.setIndent(2);
			printer.println("}");			
			printer.println("catch(mqc.exception.SystemException e)");			
			printer.println("{");			
			printer.setIndent(3);
			printer.println("System.out.println(e.toString());");						
			printer.setIndent(2);
			printer.println("}");			

			Enumeration e = iMethod.exceptionVector.elements();
			if(e.hasMoreElements())
			{
				printer.setIndent(2);
				printer.println("catch(mqc.exception.UserException e)");			
				printer.println("{");			
				printer.setIndent(3);
				printer.println("java.lang.String exceptionID = e.getID();");
				Vector elem = null;
				while(e.hasMoreElements())
				{
					elem = (Vector) e.nextElement();
					Enumeration exceptionElement = elem.elements();					
					printer.print("if(exceptionID.equals(");
					while(exceptionElement.hasMoreElements())
					{
						printer.nprint(exceptionElement.nextElement().toString());
						if(exceptionElement.hasMoreElements())
							printer.nprint(".");
					}
					printer.nprintln("Helper.id()))");
					printer.println("{");
					printer.setIndent(4);
					printer.println("mqc.holders.IntHolder ptr = new mqc.holders.IntHolder();");
					printer.println("mqc.holders.BooleanHolder bigEnd = new mqc.holders.BooleanHolder();");					
					printer.println("_msg = e.getInputStream(ptr, bigEnd);");

					exceptionElement = elem.elements();					
					printer.print("throw ");
					while(exceptionElement.hasMoreElements())
					{
						printer.nprint(exceptionElement.nextElement().toString());
						if(exceptionElement.hasMoreElements())
							printer.nprint(".");
					}
					printer.nprintln("Helper.read(_msg, ptr.value, bigEnd.value);");
					printer.setIndent(3);
					printer.println("}");
				}
				printer.println("throw new mqc.exception.UNKNOWN(\"Unknown user exception: \" + exceptionID );" );
				printer.setIndent(2);
				printer.println("}");			
			}
		}			
		if(!iMethod.returnType.equals("void"))
			printer.println("return _result;");
		pw.println("\t}");
	}

	/*********************************************************
	* Generates the skeleton code.
	* This is the server side code.
	* 
	* @param Vector The methods belonging to this interface
	*
	*********************************************************/
	public void genSkeletonCode( Vector interfaceMethods )
	{
		if(config.isInline())
			genInlinePOACode( interfaceMethods );
		else
			genPOACode( interfaceMethods );
	}


	/*********************************************************
	* Generates the skeleton code.
	* This is the server side code.
	* 
	* @param Vector The methods belonging to this interface
	*
	*********************************************************/
	public void genPOACode( Vector interfaceMethods )
	{
		PrintWriter pw = codeGenHelper.makeFile(packagePath, interfaceName + "POA");

		if(packagePath.length() != 0)
			pw.println(packageStr);
		pw.println();

		pw.print("public abstract class " + interfaceName + "POA extends mqc.Servant implements ");

		if(packageName.length() != 0)
			pw.println( packageName + "." + interfaceName + "Operations");
		else
			pw.println( interfaceName + "Operations");

		pw.println("{");
		pw.println();

		pw.println("\tpublic String getID()");
 		pw.println("\t{");
		pw.println("\t\treturn \"" + codeGenHelper.generateID(packagePath, safeName) + "\";");		
 		pw.println("\t}");
		pw.println();

		pw.println("\tpublic byte[] _invoke(Reply replyMsg, String method, byte[] buffer, mqc.holders.IntHolder ptr)");

		pw.println("\t{");
   	pw.println("\t\tmqc.CDROutputStream _output = null;");
		pw.println("\t\tmqc.CDRInputStream _input = new mqc.CDRInputStream(buffer, ptr.value, bigEndian);");

		Enumeration e = interfaceMethods.elements();
		e = interfaceMethods.elements();
		Vector params = null;
		int funNo = 0;
		InterfaceMethod iMethod = null;
		Enumeration ex = null;
		boolean first = true;

		MQCPrinter printer = new MQCPrinter();
		printer.setPW(pw);

                pw.println("");
		
		while( e.hasMoreElements() ) 
		{
			iMethod = (InterfaceMethod) e.nextElement();
			params = iMethod.paramVector;

			if(config.isCompressing())
			{
				funNo = MQCCompress.compressMethod(iMethod.name);
			}
			
			pw.println("\t\t/* " + iMethod.name + " */");
			if(first)
			{
				if(config.isCompressing())
					pw.println("\t\tif(method.equals(\"" + funNo + "\"))");
				else
			 	pw.println("\t\tif(method.equals( \"" + iMethod.name + "\"))");
				first = false;
			}
			else
			{
				if(config.isCompressing())
					pw.println("\t\telse if(method.equals(\"" + funNo + "\"))");
				else
			 	pw.println("\t\telse if(method.equals( \"" + iMethod.name + "\"))");
			}
			
			pw.println("\t\t{");
			
			printer.setIndent(3);		
			ex = iMethod.exceptionVector.elements();
			if(ex.hasMoreElements())
			{
				printer.println("try");
				printer.println("{");			
				printer.setIndent(4);
			}

			/* Print the server skeleton un-marshalling code */
			marshaller.printUnMarshal( printer, iMethod, false );
			
			if(iMethod.returnType.equals("void"))
			{
				printer.print(iMethod.name + "( ");
				codeGenHelper.printParamNames(pw, params);
				printer.nprintln(" );");
			}
			else
			{
				printer.println(codeGenHelper.mapIDL2Java( iMethod.returnType ) + " _result;");
				printer.print("_result = " + iMethod.name + "( ");
				codeGenHelper.printParamNames(pw, params);
				printer.nprintln(" );");
			}
			int bufferSize = 0;
			if(!iMethod.isOneWay)
			{
				//The exact number of bytes needed for data types itself
				bufferSize = codeGenHelper.getOutPayload(iMethod, Constants.MQC_SIZE_OF_HDR + Constants.MQC_SIZE_OF_REPLY_HDR);
				printer.println("if(handler.getProtocol().equals(\"MQCIOP\"))");
				printer.println("\tptr.value = " + (bufferSize + Constants.MQC_SIZE_OF_HDR + Constants.MQC_SIZE_OF_REPLY_HDR) + ";" );

				//printer.println("buffer = handler.createReply("+bufferSize+", ptr);");
				String bigEnd = "true";
				if(config.isForcingEndianess() == 2)
					bigEnd = "false";
				printer.println("_output = new mqc.CDROutputStream(buffer, ptr.value, " + bigEnd + ");");
			}

			/* Print the server skeleton un-marshalling code */
			marshaller.printMarshal( printer, iMethod, false );

			if(ex.hasMoreElements())
			{
				printer.setIndent(3);
				printer.println("}");
				while(ex.hasMoreElements())
				{
					String name = "";
					Enumeration exceptionElement = ((Vector) ex.nextElement()).elements();					
					while(exceptionElement.hasMoreElements())
					{
						name += exceptionElement.nextElement().toString();
						if(exceptionElement.hasMoreElements())
							name += ".";
					}
					printer.println("catch( " + name +" exception )");
					printer.println("{");
					printer.setIndent(4);
					
					String id = codeGenHelper.generateID(packagePath, name);
					int idLen = id.length() + 1;
					int index = codeGenHelper.getExPayload(Constants.MQC_SIZE_OF_HDR + Constants.MQC_SIZE_OF_REPLY_HDR, idLen, name );
					
					printer.println("if(handler.getProtocol().equals(\"MQCIOP\"))");
					printer.println("\tptr.value = " + (index + Constants.MQC_SIZE_OF_HDR + Constants.MQC_SIZE_OF_REPLY_HDR) + ";");

					printer.println("buffer = handler.createExceptionReply("+(index + Constants.MQC_SIZE_OF_HDR + 
																								 Constants.MQC_SIZE_OF_REPLY_HDR) +",ptr, "+
																								 "mqc.protocols.Message.USER_EXCEPTION);");
		
					String bigEnd = "true";
					if(config.isForcingEndianess() == 2)
						bigEnd = "false";
					printer.println("_output = new mqc.CDROutputStream(buffer, ptr.value, " + bigEnd + ");");
					printer.println(name + "Helper.write(_output, exception);");					
					printer.setIndent(3);
					printer.println("}");
				}
			}
			if(!iMethod.isOneWay)
			{
				printer.println("ptr.value = _output.getBufferPtr();");
				printer.println("return _output.getBuffer();");
			}
			else
				printer.println("return null;");

			printer.setIndent(2);
			printer.println("}");			
		}
		printer.setIndent(2);
		
		if(config.isSystemException())
			printer.println("else throw new mqc.exception.UNKNOWN(\"IDL:omg.org/CORBA/BAD_OPERATION:1.0\" );");
		else
			printer.println("return null;");
		pw.println("\t}");
		pw.println("};");
		pw.println("// End of file.");
		pw.flush();
		pw.close();
	}

	/*********************************************************
	* Generates the inline skeleton code.
	* This is the server side code. Generates support for
	* more than one protocol if neccessary
	*
	* @param Vector The methods belonging to this interface
	*********************************************************/
	public void genInlinePOACode( Vector interfaceMethods )
	{
		//Check if we have MQCIOP protocol and one or more other protocols
           /*
		Vector protocols = config.getServerProtocol();
		boolean mqcIOP = true;
		if(protocols.size() == 1)
			mqcIOP = false;

		if(mqcIOP)
		{
			mqcIOP = false;
			Enumeration e = protocols.elements();
			while(e.hasMoreElements())
			{
				if(((String) e.nextElement()).equals("MQCIOP"))
				{
					mqcIOP = true;
					break;
				}	
			}
		}
		if(!mqcIOP)
		{
			regMarshaller = marshaller;
		}
		MQCPrinter pw = genRegInlinePOACode( interfaceMethods, mqcIOP, regMarshaller );
		if(mqcIOP)
			genMQCIOPInlinePOACode( pw, interfaceMethods, marshaller );
               */

		regMarshaller = marshaller;
		MQCPrinter pw = genRegInlinePOACode( interfaceMethods, false, regMarshaller );

		pw.println("};");
		pw.println("// End of file.");
		pw.flush();
		pw.close();
	}
	
	/*********************************************************
	* Generates the inline skeleton code for the "regular"
	* protocol. This is the server side code.
	* 
	* @param Vector The methods belonging to this interface
	
	* @param boolean true if we have MQCIOP protocol, false
	*        otherwise
	* @param MQCAbstractMarshalHelper The marshaller
	*
	* @return MQCPrinter The outputstream
	*********************************************************/
	public MQCPrinter genRegInlinePOACode( Vector interfaceMethods, boolean mqcIOP, MQCAbstractMarshalHelper marshal )
	{
		PrintWriter pw = codeGenHelper.makeFile(packagePath, interfaceName + "POA");

		MQCPrinter printer = new MQCPrinter();
		printer.setPW(pw);
		printer.setIndent(3);
		
		if(packagePath.length() != 0)
			pw.println(packageStr);
		pw.println();
		pw.print("public abstract class " + interfaceName + "POA extends mqc.Servant implements ");

		if(packageName.length() != 0)
			pw.println( packageName + "." + interfaceName + "Operations");
		else
			pw.println( interfaceName + "Operations");
			
		pw.println("{");
		pw.println();

		pw.println("\tpublic String getID()");
 		pw.println("\t{");
		pw.println("\t\treturn \"" + codeGenHelper.generateID(packagePath, safeName) + "\";");		
 		pw.println("\t}");
		pw.println();
		
		pw.println("\tpublic byte[] _invoke( mqc.protocols.Reply handler, String method, byte[] _msg, mqc.holders.IntHolder ptr)");
 		pw.println("\t{");

		if(mqcIOP)
		{
			pw.println("\t\tif(handler.getProtocol().equals(\"MQCIOP\"))");
 			pw.println("\t\t\treturn invoke( handler, method, _msg, ptr);");
		}
		return printInlinePOAbody(printer, interfaceMethods, marshal);
	}

	/*********************************************************
	* Generates the inline skeleton code for the "regular"
	* protocol. This is the server side code.
	* 
	* @param MQCPrinter The outputstream
	*
	* @param Vector The methods belonging to this interface
	*
	* @param MQCAbstractMarshalHelper The marshaller
	*********************************************************/
	public void genMQCIOPInlinePOACode( MQCPrinter pw, Vector interfaceMethods, MQCAbstractMarshalHelper marshal )
	{
		pw.setIndent(1);
		
 		pw.println();
		pw.println("public byte[] invoke( mqc.protocols.Reply handler, String method, byte[] _msg, mqc.holders.IntHolder ptr)");
 		pw.println("{");

		printInlinePOAbody(pw, interfaceMethods, marshal);
	}

	/*********************************************************
	* Generates the encryption code
	* 
	* @param MQCPrinter The outputstream to write to
	*
	* @param Vector The methods
	*
	* @param MQCAbstractMarshalHelper The marshaller
	*********************************************************/
	private MQCPrinter printInlinePOAbody(MQCPrinter printer, Vector interfaceMethods, MQCAbstractMarshalHelper marshal)
	{
		Enumeration e = interfaceMethods.elements();
		InterfaceMethod iMethod = null;
		Vector params = null;
		int funNo;
		Enumeration ex = null;
		boolean first = true;
		
		while( e.hasMoreElements() ) 
		{
			printer.setIndent(2);			
			iMethod = (InterfaceMethod) e.nextElement();
			params = iMethod.paramVector;
			
			printer.println("\n\t\t/* " + iMethod.name + " */");

			funNo = MQCCompress.compressMethod(iMethod.name);
			if(first)
			{
				if(config.isCompressing())
				{
					printer.println("if(method.equals(\"" + funNo + "\"))");
				}
				else
					printer.println("if(method.equals(\"" + iMethod.name + "\"))");
				first = false;
			}
			else
			{
				if(config.isCompressing())
					printer.println("else if(method.equals(\"" + funNo + "\"))");
				else
					printer.println("else if(method.equals(\"" + iMethod.name + "\"))");
			}
			printer.println("{");

			printer.setIndent(3);			
			ex = iMethod.exceptionVector.elements();
			if(ex.hasMoreElements())
			{
				printer.println("try");
				printer.println("{");			
				printer.setIndent(4);
			}

			/* Print the server skeleton un-marshalling code */
			marshal.init();
			marshal.printUnMarshal( printer, iMethod, false );
			
			if(iMethod.returnType.equals("void"))
			{
				printer.print(iMethod.name + "( ");
				codeGenHelper.printParamNames(printer.getPW(), params);
				printer.nprintln(" );");
			}
			else
			{
				printer.println(codeGenHelper.mapIDL2Java( iMethod.returnType ) + " _result;");
				printer.print("_result = " + iMethod.name + "( ");
				codeGenHelper.printParamNames(printer.getPW(), params);
				printer.nprintln(" );");
			}

			int bufferSize = codeGenHelper.getOutPayload(iMethod, Constants.MQC_SIZE_OF_HDR + Constants.MQC_SIZE_OF_REPLY_HDR);
			if(!iMethod.isOneWay)
			{
				if(marshal.isMQCIOP())
				{
					//The exact number of bytes needed for data types itself
					bufferSize = codeGenHelper.getOutPayload(iMethod, Constants.MQC_SIZE_OF_HDR + Constants.MQC_SIZE_OF_REPLY_HDR);
					printer.println("ptr.value = " + (bufferSize + Constants.MQC_SIZE_OF_HDR + Constants.MQC_SIZE_OF_REPLY_HDR) + ";" );
				}
				//printer.println("_msg = handler.createReply("+bufferSize+",ptr);");
			}

                        printer.println( "ptr.value = 0;" );

			/* Print the server skeleton marshalling code */
			marshal.printMarshal( printer, iMethod, false );

			if(ex.hasMoreElements())
			{
				printer.setIndent(3);
				printer.println("}");

				Vector elem = null;
				while(ex.hasMoreElements())
				{
					String name = "";
					elem = (Vector) ex.nextElement();
					Enumeration exceptionElement = elem.elements();					
					while(exceptionElement.hasMoreElements())
					{
						name += exceptionElement.nextElement().toString();
						if(exceptionElement.hasMoreElements())
							name += ".";
					}
					printer.println("catch( " + name +" exception )");
					printer.println("{");
					printer.setIndent(4);
					String id = codeGenHelper.generateID(packagePath, name);
					int idLen = id.length() + 1;
					int index = codeGenHelper.getExPayload(Constants.MQC_SIZE_OF_HDR + Constants.MQC_SIZE_OF_REPLY_HDR, idLen, name );
					printer.println("ptr.value = " + (index + Constants.MQC_SIZE_OF_HDR + Constants.MQC_SIZE_OF_REPLY_HDR) + ";");
					printer.println("_msg = handler.createExceptionReply("+bufferSize+", ptr, mqc.protocols.Message.USER_EXCEPTION);");
					printer.println("_msg = " + name + "Helper.write(_msg, ptr, exception);");
					printer.setIndent(3);
					printer.println("}");
				}
			}
			if(!iMethod.isOneWay)
			{
				printer.println("return _msg;");
			}
			else
				printer.println("return null;");
				
			printer.setIndent(2);
			printer.println("}");
		}
		if(config.isSystemException())
			printer.println("else throw new mqc.exception.UNKNOWN(\"Unknown method: \" + method );");
		else
			printer.println("return null;");

		printer.setIndent(1);
		printer.println("}");
		printer.setIndent(0);
		return printer;
	}
		
}

