/*
 * Copyright (c) 2003 David E. Bakken, his research students, and Washington State University.
 * Please see the file LICENSE.pdf for more details on terms of use.
 */

package mqc.protocols;
import mqc.*;

/**
 * Class representing a GIOPReply Message
 *  
 * @author Olav Haugan
 * @author Eivind Naess
 * @author Kim Christian Swenson
 *
 * struct GIOPLiteReplyHeader {
 *    unsigned long request_id;
 *    unsigned long reply_status;
 * };
 *
 */
public class GIOPLiteReply implements GIOPLiteMessage, Reply
{
   protected GIOPLiteHeader m_GIOPLiteHeader;
   
   /** The reply status field from this message */
   protected int m_replyStatus;

   /** The request id field for this message */
   protected int m_reqID;

   /** Size of the reply msg */
   public static final int SIZE_OF_REP = 8;

   protected Body m_body;

   public GIOPLiteReply()
   {
      this( new GIOPLiteHeader( GIOPLiteMessage.REPLY_MSG ) );
   }

   public GIOPLiteReply( GIOPLiteHeader header )
   {
      m_GIOPLiteHeader = header;
      m_body = new Body();
   }

   public GIOPLiteHeader getGIOPLiteHeader()
   {
      return m_GIOPLiteHeader;
   }

   public void setGIOPLiteHeader( GIOPLiteHeader header )
   {
      m_GIOPLiteHeader = header;
   }

   public Body getBody()
   {
      return m_body;
   }

   public void setBody( Body body )
   {
      m_body = body;
   }

   /**
    * Return the Reply Status field of the demarshalled reply messages
    * @return Reply the deMarshalled message's reply status
    */
   public int getReplyStatus()
   {
      return m_replyStatus;
   }

   public void setReplyStatus( int status )
   {
      m_replyStatus = status;
   }

   /**
    * Returns the Request ID for this message 
    */
   public int getRequestID()
   {
      return m_reqID;
   }

   /**
    * Sets the message's request ID
    * @param reqId The request ID
    */
   public void setRequestID(int reqID)
   {
      m_reqID = reqID;
   }

   public int getCompletionStatus()
   {
      return m_body.getCompletionStatus();
   }

   public void setCompletionStatus( int status )
   {
      m_body.setCompletionStatus( status );
   }

   public int getExceptionMinorCode()
   {
      return m_body.getExceptionMinorCode();
   }

   public void setExceptionMinorCode( int minor )
   {
      m_body.setExceptionMinorCode( minor );
   }

   public int getExceptionMinorVMCID()
   {
      return m_body.getExceptionMinorVMCID();
   }

   public void setExceptionMinorVMCID( int vmcid )
   {
      m_body.setExceptionMinorVMCID( vmcid );
   }

   public String getExceptionID()
   {
      return m_body.getExceptionID();
   }

   public void setExceptionID( String id )
   {
      m_body.setExceptionID( id );
   }

   public boolean isBigEndian()
   {
      return !(Config.FORCE_ENDIANESS && !Config.BIG_ENDIAN);
   }

   public void setBigEndian( boolean bigEndian )
   {
   }

   public int getHdrSize()
   {
      return 8;
   }

   /**
    * Demarshal a giop lite reply message 
    */
   public void deMarshal( CDRInputStream in )
   {
      /* TODO implement Exception handling */

      in.reset();
      in.setBigEndian( isBigEndian() );

      // GIOPLite header
      m_GIOPLiteHeader.deMarshal( in );

      // unsigned long request_id
      m_reqID = in.readUnsignedLong();

      // unsigned long reply_status
      m_replyStatus = in.readUnsignedLong();

      // GIOPLite Request Body

      // Copy data from underlying buffer of 
      // CDRInputStream into body buffer, and adjust
      // the Stream's pointer acordingly.
      
      // Get bufferPointer and align it with 4-bytes
      int bufPtr = in.bytesStreamed();
      bufPtr = (bufPtr + 3) & ~3;

      // Make sure message body's buffer is big enough
      m_body.setMinSize( m_GIOPLiteHeader.getSize() - bufPtr + getHdrSize() );

      // copy body part of message from inputstream into body
      System.arraycopy( in.getBuffer(), bufPtr, m_body.getBuffer(), 0, m_GIOPLiteHeader.getSize() - bufPtr + getHdrSize() );

      // Adjust buffer pointer
      in.setPtr( m_GIOPLiteHeader.getSize() + getHdrSize() );
   }

   public void marshal ( CDROutputStream out )
   {
      /* TODO implement Exception handling */

      out.reset();
      out.setBigEndian( isBigEndian() );

      // GIOPLite header
      m_GIOPLiteHeader.marshal( out );

      // unsigned long request_id
      out.writeUnsignedLong( m_reqID );

      // unsigned long reply_status
      out.writeUnsignedLong( m_replyStatus );

      // GIOPLite Request Body

      // Copy data from body buffer into underlying
      // buffer of CDROutputStream, and adjust
      // the Stream's pointer acordingly.

      // get underlying buffer
      byte[] buf = out.getBuffer();
      
      // Get bufferPointer and align it with 4-bytes
      int bufPtr = out.bytesStreamed();
      bufPtr = (bufPtr + 3) & ~3;

      // Hack ... set message size again now that we know it .. !
      m_GIOPLiteHeader.setSize( bufPtr - getHdrSize() + m_body.getPtr() );
      GIOPLiteHeader.setMsgSize( buf, m_GIOPLiteHeader.getSize(), isBigEndian() );

      // Resize if necessary
      if ( buf.length < bufPtr + m_body.getPtr() )
      {
         byte[] tmp = new byte[bufPtr + m_body.getPtr()];
         System.arraycopy(buf, 0, tmp, 0, out.bytesStreamed() );
         buf = tmp;
      }

      // copy data
      System.arraycopy( m_body.getBuffer(), 0, buf, bufPtr, m_body.getPtr() );

      // Insert underlying buffer in case of resize
      out.setBuffer( buf );

      // Adjust buffer pointer
      out.setPtr( bufPtr + m_body.getPtr() );
   }

}
