/*
 * Copyright (c) 2003 David E. Bakken, his research students, and Washington State University.
 * Please see the file LICENSE.pdf for more details on terms of use.
 */

package mqc.protocols;

public class GIOPReply implements GIOPMessage, Reply
{
   protected GIOPHeader m_GIOPHeader;
   protected GIOPReplyHeader m_GIOPReplyHeader;
   protected Body m_body;

   public GIOPReply( GIOPHeader header, GIOPReplyHeader repHeader )
   {
      m_GIOPHeader = header;
      m_GIOPHeader.setMessageType( GIOPMessage.REPLY_MSG );
      m_GIOPReplyHeader = repHeader;
      m_body = new Body();
   }

   public GIOPHeader getGIOPHeader()
   {
      return m_GIOPHeader;
   }

   public void setGIOPHeader( GIOPHeader header )
   {
      m_GIOPHeader = header;
   }

   public GIOPReplyHeader getGIOPReplyHeader()
   {
      return m_GIOPReplyHeader;
   }

   public void setGIOPReplyHeader( GIOPReplyHeader header )
   {
      m_GIOPReplyHeader = header;
   }

   public Body getBody()
   {
      return m_body;
   }

   public void setBody( Body body )
   {
      m_body = body;
   }

   public int getReplyStatus()
   {
      return m_GIOPReplyHeader.getReplyStatus();
   }

   public void setReplyStatus( int status )
   {
      m_GIOPReplyHeader.setReplyStatus( status );
   }

   public int getCompletionStatus()
   {
      return m_body.getCompletionStatus();
   }

   public void setCompletionStatus( int status )
   {
      m_body.setCompletionStatus( status );
   }

   public int getExceptionMinorCode()
   {
      return m_body.getExceptionMinorCode();
   }

   public void setExceptionMinorCode( int minor )
   {
      m_body.setExceptionMinorCode( minor );
   }

   public int getExceptionMinorVMCID()
   {
      return m_body.getExceptionMinorVMCID();
   }

   public void setExceptionMinorVMCID( int vmcid )
   {
      m_body.setExceptionMinorVMCID( vmcid );
   }

   public String getExceptionID()
   {
      return m_body.getExceptionID();
   }

   public void setExceptionID( String id )
   {
      m_body.setExceptionID( id );
   }

   public int getRequestID()
   {
      return m_GIOPReplyHeader.getRequestID();
   }

   public void setRequestID( int reqID )
   {
      m_GIOPReplyHeader.setRequestID( reqID );
   }

   public void setBigEndian( boolean bigEndian)
   {
      m_GIOPHeader.setBigEndian( bigEndian );
   }

   public boolean isBigEndian()
   {
      return m_GIOPHeader.isBigEndian();
   }

   public int getHdrSize()
   {
      return 12;
   }

   public void deMarshal( CDRInputStream in ) throws RuntimeException
   {
      //Reset stream
      in.reset();

      // GIOP header
      m_GIOPHeader.deMarshal( in );

      // GIOP Reply header
      m_GIOPReplyHeader.deMarshal( in );

      // GIOP Reply Body

      // Copy data from underlying buffer of 
      // CDRInputStream into body buffer, and adjust
      // the Stream's pointer acordingly.

      // Get bufferPointer
      int bufPtr = in.bytesStreamed();

      // If GIOP minor version >= 2, align it with 8-bytes
      if ( m_GIOPHeader.getMinor() >= 2 )
      {
         bufPtr = (bufPtr + 7) & ~7;
         in.setPtr( bufPtr );
      }

      /* NO_EXCEPTION */
      if ( m_GIOPReplyHeader.getReplyStatus() == Reply.NO_EXCEPTION )
      {
         // Make sure message body's buffer is big enough
         m_body.setMinSize( m_GIOPHeader.getMessageSize() - bufPtr + getHdrSize() );

         // copy body part of message from inputstream into body
         System.arraycopy( in.getBuffer(), bufPtr, m_body.getBuffer(), 0, m_GIOPHeader.getMessageSize() - bufPtr + getHdrSize() );

         // Adjust buffer pointer
         in.setPtr( m_GIOPHeader.getMessageSize() + getHdrSize() );
      }

      /* USER_EXCEPTION */
      else if ( m_GIOPReplyHeader.getReplyStatus() == Reply.USER_EXCEPTION )
      {
         /* TODO implement user exception */
         setExceptionID( in.readString() );
         setExceptionMinorCode( in.readUnsignedLong() );
         setCompletionStatus( in.readUnsignedLong() );
      }

      /* SYSTEM_EXCEPTION */
      else if ( m_GIOPReplyHeader.getReplyStatus() == Reply.SYSTEM_EXCEPTION )
      {
         /* TODO implement non standard system exception */
         setExceptionID( in.readString() );
         setExceptionMinorCode( in.readUnsignedLong() );
         setCompletionStatus( in.readUnsignedLong() );
      }

   }

   public void marshal( CDROutputStream out )
   {
      //Reset stream
      out.reset();

      // GIOP header
      m_GIOPHeader.marshal( out );

      // GIOP Reply header
      m_GIOPReplyHeader.marshal( out );

      // GIOP Reply Body

      // Copy data from body buffer into underlying
      // buffer of CDROutputStream, and adjust
      // the Stream's pointer acordingly.

      // get underlying buffer
      byte[] buf = out.getBuffer();
      
      // Get bufferPointer
      int bufPtr = out.bytesStreamed();

      // If GIOP minor version >= 2, align it with 8-bytes
      if ( m_GIOPHeader.getMinor() >= 2 )
      {
         bufPtr = (bufPtr + 7) & ~7;
         out.setPtr( bufPtr );
      }

      /* NO_EXCEPTION */
      if ( m_GIOPReplyHeader.getReplyStatus() == Reply.NO_EXCEPTION )
      {
         // Resize if necessary
         if ( buf.length < bufPtr + m_body.getPtr() )
         {
            byte[] tmp = new byte[bufPtr + m_body.getPtr()];
            System.arraycopy(buf, 0, tmp, 0, out.bytesStreamed() );
            buf = tmp;
         }
   
         // copy data
         System.arraycopy( m_body.getBuffer(), 0, buf, bufPtr, m_body.getPtr() );
   
         // Insert underlying buffer in case of resize
         out.setBuffer( buf );
   
         // Adjust buffer pointer
         out.setPtr( bufPtr + m_body.getPtr() );
      }

      /* USER_EXCEPTION */
      else if ( m_GIOPReplyHeader.getReplyStatus() == Reply.USER_EXCEPTION )
      {
         /* TODO implement user exception */
         out.writeString( getExceptionID() );
         out.writeUnsignedLong( getExceptionMinorCode() );
         out.writeUnsignedLong( getCompletionStatus() );
      }

      /* SYSTEM_EXCEPTION */
      else if ( m_GIOPReplyHeader.getReplyStatus() == Reply.SYSTEM_EXCEPTION )
      {
         /* TODO implement non standard system exception */
         out.writeString( getExceptionID() );
         out.writeUnsignedLong( getExceptionMinorCode() );
         out.writeUnsignedLong( getCompletionStatus() );
      }

      // Hack ... set message size again now that we know it .. !
      m_GIOPHeader.setMessageSize( out.bytesStreamed() );
      GIOPHeader.setMsgSize( buf, m_GIOPHeader.getMessageSize(), m_GIOPHeader.isBigEndian() );

   }

}
