/*
 * Copyright (c) 2003 David E. Bakken, his research students, and Washington State University.
 * Please see the file LICENSE.pdf for more details on terms of use.
 */

//Use MQC package only to get to work with two different
//libraries

package mqc;

/**
* Class for CDR input stream. Un-marshals the data in
* the buffer.
* 
* @author Olav Haugan
*
* 14.10.01
*/
public class CDRInputStream
{
   private boolean bigEndian;
   private byte[] buffer;
   private int bufferPtr;

  /**
   * Constructor. Initializes variables.
   */
   public CDRInputStream()
   {
      buffer = null;
      bufferPtr = 0;
      bigEndian = true;
   }

  /**
   * Constructor. Initializes variables.
   *
   * @param byte[] The buffer to stream from
   * @param int The buffer pointer
   * @param boolean true if big endian, false otherwise
   */
   public CDRInputStream(byte[] buffer, int bufferPtr, boolean bigEndian)
   {
      this.buffer = buffer;
      this.bufferPtr = bufferPtr;
      this.bigEndian = bigEndian;
   }

  /**
   * Gets the buffer
   *
   * @return byte[] The buffer
   */
   public boolean getBigEndian()
   {
      return bigEndian;
   }

  /**
   * Un-marshals a byte from the stream.
   *
   * @return byte a byte
   */
   public byte read_octet()
   {
      return (byte) buffer[bufferPtr++];
   }

  /**
   * Un-marshals a boolean from the stream.
   *
   * @return boolean A boolean
   */
   public boolean read_boolean()
   {
      return  ((buffer[bufferPtr++] == 1) ? true : false);
   }

  /**
   * Un-marshals a char from the stream.
   *
   * @return char A char
   */
   public char read_char()
   {
      return (char) buffer[bufferPtr++];
   }

  /**
   * Un-marshals an unsigned short from the stream.
   *
   * @return unsigned short An unsigned short
   */
   public short read_ushort()
   {
      return read_short();
   }

  /**
   * Un-marshals an short from the stream.
   *
   * @return short An unsigned short
   */
   public short read_short()
   {
      bufferPtr = (bufferPtr +1) & ~1;
      if(bigEndian)
      {
         return (short) (((buffer[bufferPtr++] & 0xFF) << 8) |
                         (buffer[bufferPtr++] & 0xFF));
      }
      else
      {
         return (short) ((buffer[bufferPtr++] & 0xFF) |
                         ((buffer[bufferPtr++] & 0xFF) << 8));
      }
   }

  /**
   * Un-marshals an unsigned long from the stream.
   *
   * @return unsigned long An unsigned long
   */
   public int read_ulong()
   {
      return read_long();
   }

  /**
   * Un-marshals an long from the stream.
   *
   * @return long An unsigned long
   */
   public int read_long()
   {
      bufferPtr = (bufferPtr +3) & ~3;
      if(bigEndian)
      {
         return (int) (((buffer[bufferPtr++] & 0xFF) << 24) |
                       ((buffer[bufferPtr++] & 0xFF) << 16) |
                       ((buffer[bufferPtr++] & 0xFF) << 8) |
                       (buffer[bufferPtr++] & 0xFF));
      }
      else
      {
         return (int) ((buffer[bufferPtr++] & 0xFF) |
                       ((buffer[bufferPtr++] & 0xFF) << 8) |
                       ((buffer[bufferPtr++] & 0xFF) << 16) |
                       ((buffer[bufferPtr++] & 0xFF) << 24));
      }
   }

   /**
   * Un-marshals an unsigned long long from the stream.
   *
   * @return unsigned long long An unsigned long long
   */
   public long read_ulonglong()
   {
      return read_longlong();
   }

  /**
   * Un-marshals a long long from the stream
   *
   * @return long long A long long
   */
   public long read_longlong()
   {
      long val = 0;
      long lval1 = 0;
      long lval2 = 0;

      bufferPtr = (bufferPtr + 7) & ~7;
      if(bigEndian)
      {
         lval1 = (long) (((buffer[bufferPtr++] & 0xFF) << 24) |
                         ((buffer[bufferPtr++] & 0xFF) << 16) |
                         ((buffer[bufferPtr++] & 0xFF) <<  8) |
                         (buffer[bufferPtr++] & 0xFF));

         lval2 = (long) (((buffer[bufferPtr++] & 0xFF) << 24) |
                         ((buffer[bufferPtr++] & 0xFF) << 16) |
                         ((buffer[bufferPtr++] & 0xFF) << 8) |
                         (buffer[bufferPtr++] & 0xFF));

         val = (long) ((lval1 << 32) | (lval2 & 0xFFFFFFFFl));
      }
      else
      {
         lval1 = (long) (((buffer[bufferPtr++] & 0xFF) << 0) |
                         ((buffer[bufferPtr++] & 0xFF) << 8) |
                         ((buffer[bufferPtr++] & 0xFF) << 16) |
                         ((buffer[bufferPtr++] & 0xFF)  << 24));

         lval2 = (long) (((buffer[bufferPtr++] & 0xFF) << 0) |
                         ((buffer[bufferPtr++] & 0xFF) << 8) |
                         ((buffer[bufferPtr++] & 0xFF) << 16) |
                         ((buffer[bufferPtr++] & 0xFF)  << 24));

         val = (long) ((lval2 << 32) | (lval1 & 0xFFFFFFFFl));
      }
      return val;
   }

  /**
   * Un-marshals a float from the stream
   *
   * @return float A float
   */
   public float read_float()
   {
      return java.lang.Float.intBitsToFloat(read_long());
   }

  /**
   * Un-marshals a double from the stream
   *
   * @return double A double
   */
   public double read_double()
   {
      return java.lang.Double.longBitsToDouble(read_longlong());
   }

   /**
   * Un-marshals a string from the stream
   *
   * @return String A string
   */
   public String read_string()
   {
      int len = read_long();
      len--;
      byte[] str = new byte[len];
      System.arraycopy(buffer, bufferPtr, str, 0, len);
      bufferPtr += len + 1;
      return new String(str);
   }

   /**
   * Sets the buffer
   *
   * @param byte[] The buffer
   *
   */
   public void setBuffer(byte[] buffer)
   {
      this.buffer = buffer;
   }

   /**
   * Gets the buffer
   *
   * @return byte[] The buffer
   *
   */
   public byte[] getBuffer()
   {
      return buffer;
   }

   /**
   * Gets the buffer pointer
   *
   * @return int The buffer pointer
   *
   */
   public int getBufferPtr()
   {
      return bufferPtr;
   }
}
